use binrw::{binrw, BinRead, BinWrite};
use std::{
    mem::size_of,
    ops::{Deref, DerefMut},
};

/// Wrapper to prefix arbitrary structs with size of struct as u32
#[binrw]
#[derive(Debug, PartialEq)]
pub struct FieldOf<T>
where
    T: for<'a> BinRead<Args<'a> = ()>,
    T: for<'a> BinWrite<Args<'a> = ()>,
{
    // Temporary variable that must match size of value
    #[br(temp, assert(size == size_of::<T>().try_into().unwrap()))]
    #[bw(try_calc(u32::try_from(size_of::<T>())))]
    size: u32,

    pub value: T,
}

impl<T> Deref for FieldOf<T>
where
    T: for<'a> BinRead<Args<'a> = ()>,
    T: for<'a> BinWrite<Args<'a> = ()>,
{
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.value
    }
}

impl<T> DerefMut for FieldOf<T>
where
    T: for<'a> BinRead<Args<'a> = ()>,
    T: for<'a> BinWrite<Args<'a> = ()>,
{
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.value
    }
}
